/* drivers/cci/sensors/capella/capella_cm3623.c
 *
 * Copyright (C) 2011 Compalcomm
 * Author: Ron Kuo <ron_kuo@compalcomm.com>
 *
 * Revise List:
 * 1. the first version in 2011/04/20
 * 2. add early_suspend and late_resume func in 2011/04/27
 * 3. modification of suspend and resume for p-sensor,
 *    early_suspend and late_resume func for light sensor
 *    adc_to_distance table of p-sensor in 2011/05/03.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/irq.h>
#include <linux/input.h>
#include <linux/platform_device.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/miscdevice.h>
#include <linux/workqueue.h>
#include <linux/mutex.h>
#include <linux/err.h>
#include <linux/i2c.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/gpio.h>
//#include "mach/../../cci_smem.h"
#include "mach/../../smd_private.h"
#include <mach/vreg.h>
#include <mach/mpp.h>

#ifdef CONFIG_HAS_EARLYSUSPEND
#include <linux/earlysuspend.h>
#endif

#define CM3623_DEFAULT_DELAY   100
#define CM3623_USE_INTERRUPT   0 
#define CM3623_INT       124
/* Defines */
#if CM3623_USE_INTERRUPT
#define CM3623_INIT_VALUE   0x10
#else
#define CM3623_INIT_VALUE   0x20
#endif

#define CM3623_INITIAL_I2C_ADDR     (0x22 >> 1)
#define CM3623_ALS_I2C_ADDR         (0x20 >> 1)
#define CM3623_PS_I2C_ADDR          (0xb0 >> 1)
#define CM3623_I2C_THD_ADDR         (0xb2 >> 1)
#define CM3623_PS_DATA_I2C_ADDR     (0xb1 >> 1)
#define CM3623_ALS_I2C_DATA_LSB     ((0x23-1) >> 1)
#define CM3623_ALS_I2C_DATA_MSB     ((0x21-1) >> 1)

/* ALS */
#define CM3623_SD_ALS       0x01        // ALS Shutdown Mode Setting
#define CM3623_WDM          0x02        // ALS Data Mode Setting : byte or word
#define CM3623_IT_ALS_100MS (0x00 << 2) // ALS Integration Time 100ms
#define CM3623_IT_ALS_200MS (0x01 << 2) // ALS Integration Time 200ms
#define CM3623_IT_ALS_400MS (0x02 << 2) // ALS Integration Time 400ms
#define CM3623_IT_ALS_800MS (0x03 << 2) // ALS Integration Time 800ms
#define CM3623_THD_ALS_0    (0x00 << 4) // ALS Threshold Window Setting
#define CM3623_THD_ALS_1    (0x01 << 4) // ALS Threshold Window Setting
#define CM3623_THD_ALS_2    (0x02 << 4) // ALS Threshold Window Setting
#define CM3623_THD_ALS_3    (0x03 << 4) // ALS Threshold Window Setting
#define CM3623_GAIN_ALS_1   (0x00 << 6) // ALS Gain setting for Sensitivity Range Selection (divide by 1)
#define CM3623_GAIN_ALS_2   (0x01 << 6) // ALS Gain setting for Sensitivity Range Selection (divide by 2)
#define CM3623_GAIN_ALS_4   (0x02 << 6) // ALS Gain setting for Sensitivity Range Selection (divide by 4)
#define CM3623_GAIN_ALS_8   (0x03 << 6) // ALS Gain setting for Sensitivity Range Selection (devide by 8)
/* PS */
#define CM3623_SD_PS        0x01        // PS Shutdown Mode Setting
#define CM3623_RESERVE_PS   0x02        // PS Reserve Bit
#define CM3623_INT_PS       0x04        // PS Interruption Setting
#define CM3623_INT_ALS      0x08        // ALS Interruption Setting
#define CM3623_IT_PS_0      (0x00 << 4) // PS Integration Time 1T
#define CM3623_IT_PS_1      (0x01 << 4) // PS Integration Time 1.1875T
#define CM3623_IT_PS_2      (0x02 << 4) // PS Integration Time 1.4375T
#define CM3623_IT_PS_3      (0x03 << 4) // PS Integration Time 1.875T
#define CM3623_DR_PS_160    (0x00 << 6) // IR LED On/Off Duty Ratio Setting (1:160)
#define CM3623_DR_PS_320    (0x01 << 6) // IR LED On/Off Duty Ratio Setting (1:320)
#define CM3623_DR_PS_640    (0x02 << 6) // IR LED On/Off Duty Ratio Setting (1:640)
#define CM3623_DR_PS_1280   (0x03 << 6) // IR LED On/Off Duty Ratio Setting (1:1280)

/* Initial value for ALS write register and PS write register */
#define CM3623_ALS_WRITE_REG        (CM3623_GAIN_ALS_1 | CM3623_THD_ALS_1 | CM3623_IT_ALS_400MS | CM3623_WDM)
#if CM3623_USE_INTERRUPT
#define CM3623_PS_WRITE_REG         (CM3623_DR_PS_160 | CM3623_IT_PS_3 | CM3623_INT_PS | CM3623_INT_ALS)
#else
#define CM3623_PS_WRITE_REG         (CM3623_DR_PS_160 | CM3623_IT_PS_3)
#endif
#define CM3623_LIGHT_OFF            (CM3623_ALS_WRITE_REG | CM3623_SD_ALS)
#define CM3623_LIGHT_ON             (CM3623_ALS_WRITE_REG & (~CM3623_SD_ALS))
#define CM3623_PROXIMITY_OFF        (CM3623_PS_WRITE_REG | CM3623_SD_PS)
#define CM3623_PROXIMITY_ON         (CM3623_PS_WRITE_REG & (~CM3623_SD_PS))

struct capella_cm3623_data {
    struct i2c_client *client;
    struct work_struct sensor_work;

    struct input_dev *als_input_dev;
    atomic_t als_enable;
    atomic_t als_delay;
    uint16_t als_last;
    struct delayed_work als_work;

    struct input_dev *ps_input_dev;
    atomic_t ps_enable;
    atomic_t ps_delay;
    uint16_t ps_last;
    struct mutex enable_mutex;
    struct mutex data_mutex;
    struct delayed_work ps_work;
};

//  global structure
static struct capella_cm3623_data *this_data;
static bool cm3623_light_status = false;
static bool cm3623_proximity_status = false;
static uint16_t PROXIMITY_NV_ITEM_VALUE = 3;
static uint16_t PROXIMITY_NV_ITEM_VALUE_HIGH = 5;
const int LIGHT_LUX_BENCHMARK = 500;
static uint16_t LIGHT_NV_ITEM_VALUE = 250;
static uint16_t LIGHT_NV_ITEM_VALUE_HIGH = 360;
static uint16_t LIGHT_NV_ITEM_VALUE_LOW = 190;
#ifdef CONFIG_HAS_EARLYSUSPEND
static struct early_suspend early_suspend;
#endif

#define CM3623_DEBUG (0)
#if CM3623_DEBUG
#define D(x...) pr_info(x)
#else
#define D(x...)
#endif

/*
 * client: target client
 * buf: target register with command
 * count: length of transmitting
 */
static int i2c_write(struct i2c_client *client, char *buf, int count)
{
    int ret;
    ret = i2c_master_send(client, buf, count);
    if (ret != count) {
        printk(KERN_INFO "[CM3623] : i2c_write --> Send reg. info error \n");
        printk(KERN_INFO "[CM3623] : i2c_write --> Err Code = %d\n", ret);
        return -1;
    }
    return 0;
}

static int cm3623_suspend(struct i2c_client *client, pm_message_t mesg)
{
    char reg_buff[1] = {0};
    int rs = 0;

    client = this_data->client;
    printk(KERN_INFO "[CM3623] : %s\n", __func__);

    if (cm3623_proximity_status == true) {
        if (atomic_cmpxchg(&this_data->ps_enable, 1, 0)) {
            cancel_delayed_work_sync(&this_data->ps_work);
            atomic_set(&this_data->ps_enable, 0);

            mutex_lock(&this_data->enable_mutex);
            client->addr = CM3623_PS_I2C_ADDR;
            reg_buff[0] = CM3623_PROXIMITY_OFF;     // turn off proximity sensor
            rs = i2c_write(client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : Turn off proximity sensor => [FAIL]\n");
            }
            else {
                D("[CM3623] : Enter suspend and turn off proximity sensor => [SUCCEED]\n");
            }
            mutex_unlock(&this_data->enable_mutex);
        }
    }

    return 0;
}

static int cm3623_resume(struct i2c_client *client)
{
    //struct i2c_client *client = this_data->client;
    char reg_buff[1] = {0};
    int rs = 0;

    client = this_data->client;
    printk(KERN_INFO "[CM3623] : %s\n", __func__);

    if (cm3623_proximity_status == true) {
        if (!atomic_cmpxchg(&this_data->ps_enable, 0, 1)) {
            unsigned int delay = atomic_read(&this_data->ps_delay);

            mutex_lock(&this_data->enable_mutex);
            client->addr = CM3623_PS_I2C_ADDR;
            reg_buff[0] = CM3623_PROXIMITY_ON;     // turn on proximity sensor
            rs = i2c_write(client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : Turn on proximity sensor setting => [FAIL]\n");
            }
            else {
                schedule_delayed_work(&this_data->ps_work, msecs_to_jiffies(delay));
                atomic_set(&this_data->ps_enable, 1);
                D("[CM3623] : Enter resume and turn on proximity sensor => [SUCCEED]\n");
            }
            mutex_unlock(&this_data->enable_mutex);
        }
    }

    return 0;
}

#ifdef CONFIG_HAS_EARLYSUSPEND
void cm3623_light_early_suspend(struct early_suspend *h)
{
    struct i2c_client *client = this_data->client;
    char reg_buff[1] = {0};
    int rs = 0;

    printk(KERN_INFO "[CM3623] : %s\n", __func__);
    if (cm3623_light_status == true) {
        if (atomic_cmpxchg(&this_data->als_enable, 1, 0)) {
            cancel_delayed_work_sync(&this_data->als_work);
            atomic_set(&this_data->als_enable, 0);

            mutex_lock(&this_data->enable_mutex);
            client->addr = CM3623_ALS_I2C_ADDR;
            reg_buff[0] = CM3623_LIGHT_OFF;   // turn off ambient light sensor
            rs = i2c_write(client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : Turn off light sensor setting => [FAIL]\n");
            }
            else {
                D("[CM3623] : Enter early-suspend and turn off light sensor => [SUCCEED]\n");
            }
            mutex_unlock(&this_data->enable_mutex);
        }
    }
}

void cm3623_light_late_resume(struct early_suspend *h)
{
    struct i2c_client *client = this_data->client;
    char reg_buff[1] = {0};
    int rs = 0;

    printk(KERN_INFO "[CM3623] : %s\n", __func__);
    if (cm3623_light_status == true) {
        if (!atomic_cmpxchg(&this_data->als_enable, 0, 1)) {
            unsigned int delay = atomic_read(&this_data->als_delay);

            mutex_lock(&this_data->enable_mutex);
            client->addr = CM3623_ALS_I2C_ADDR;
            reg_buff[0] = CM3623_LIGHT_ON;   // turn on ambient light sensor
            rs = i2c_write(client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : Turn on light sensor setting => [FAIL]\n");
            }
            else {
                schedule_delayed_work(&this_data->als_work, msecs_to_jiffies(delay));
                atomic_set(&this_data->als_enable, 1);
                D("[CM3623] : Enter late-resume and turn on light sensor => [SUCCEED]\n");
            }
            mutex_unlock(&this_data->enable_mutex);
        }
    }
}
#endif

int report_ps_data(void)
{
    struct i2c_client *client;
    unsigned char reg_buf[1] = {0};
    int rs=0;

    uint16_t ps_value_tmp = 0;
    //static uint16_t old_ps_adc = 0;
  
    D("[CM3623] : -----------------------------\n"); 
    D("[CM3623] : %s\n Begin", __func__);
    client =this_data->client;
    client->addr = CM3623_PS_DATA_I2C_ADDR;
    rs = i2c_master_recv(client, reg_buf, 1);
    D("[CM3623] : rs = %d\n ", rs);
    if (rs != 1) {
        printk(KERN_INFO "[CM3623] : i2c_read ps data fial\n");
        return rs;
    } 
    else {
        ps_value_tmp = reg_buf[0];                          
        D("[CM3623] : Proximity-sensor value = %d\n", ps_value_tmp);

        this_data->ps_last = ps_value_tmp;

        D("[CM3623] : get_proximity_data = %d\n", ps_value_tmp);
        if (ps_value_tmp > PROXIMITY_NV_ITEM_VALUE) {
            input_report_abs(this_data->ps_input_dev, ABS_X, 0);
            input_sync(this_data->ps_input_dev);
        }
        else {
            input_report_abs(this_data->ps_input_dev, ABS_X, 1);
            input_sync(this_data->ps_input_dev);
        }

        D("[CM3623] : %s\n Finish", __func__);
        D("[CM3623] : -----------------------------\n"); 
        return 0 ;
    }

}

#define ALS_CHANGE_THRESHOLD 2
int report_als_data(void)
{   
    struct i2c_client *client;
    unsigned char lsb[1], msb[1];
    int rs=0;
    uint16_t als_value_tmp = 0;
    uint16_t lux_last;
    
    D("[CM3623] : -----------------------------\n"); 
    D("[CM3623] : %s\n Begin", __func__);
    client =this_data->client;
    memset(lsb, 0, 1);
    memset(msb, 0, 1);

    client->addr = CM3623_ALS_I2C_DATA_LSB;
    rs = i2c_master_recv(client, lsb, 1);
    if (rs != 1) {
        printk(KERN_INFO "[CM3623] : i2c_read lsb data fail in report_als_data()\n");
        return rs;
    } 

    client->addr = CM3623_ALS_I2C_DATA_MSB;
    rs = i2c_master_recv(client, msb, 1);
    if (rs != 1) {
        printk(KERN_INFO "[CM3623] : i2c_read msb data fail in report_als_data()\n");
        return rs;
    } 

    lux_last = this_data->als_last;
    als_value_tmp = (uint16_t)((((msb[0] * 256) + lsb[0]) * LIGHT_LUX_BENCHMARK)/LIGHT_NV_ITEM_VALUE);    // under IT Timeing = 400ms ; 1adc * 3 ~ 1lux
    D("[CM3623] : Light sensor lux= %d\n", als_value_tmp);
 
    if (unlikely(abs(als_value_tmp - lux_last) >= ALS_CHANGE_THRESHOLD))
    {
        this_data->als_last = als_value_tmp;
        D("[CM3623] : Report data to input sub-system\n");
        input_report_abs(this_data->als_input_dev, ABS_X, als_value_tmp); // notices input sub-system that gets value
        input_sync(this_data->als_input_dev);                         // wait till input sub-system finishs job.
    }

    D("[CM3623] : %s\n Finish", __func__);
    D("[CM3623] : -----------------------------\n");
    return 0;
}

// delay work for ps 根據delay時間去告知過多久才回報資料
static void ps_work_func(struct work_struct *work)
{
    struct capella_cm3623_data *cm3623_ptr;
    unsigned int delay;
    int ret;

    D("[CM3623] : %s\n", __func__);
    cm3623_ptr = container_of((struct delayed_work *)work, struct capella_cm3623_data, ps_work);
    delay = atomic_read(&cm3623_ptr->ps_delay);

    ret = report_ps_data();

    if (ret != 0) {
        printk(KERN_INFO "[CM3623] : report_ps_data fail\n");
    }

    schedule_delayed_work(&cm3623_ptr->ps_work,msecs_to_jiffies(delay));
}


// delay work for als 根據delay時間去告知過多久才回報資料
static void als_work_func(struct work_struct *work)
{
    struct capella_cm3623_data *cm3623_ptr;
    unsigned int delay;
    int ret;

    D("[CM3623] : %s\n", __func__);
    cm3623_ptr = container_of((struct delayed_work *)work, struct capella_cm3623_data, als_work);
    delay = atomic_read(&cm3623_ptr->als_delay);

    ret = report_als_data();

    if (ret != 0) {
        printk(KERN_INFO "[CM3623] : report_als_data fail\n");
    }

    schedule_delayed_work(&cm3623_ptr->als_work,msecs_to_jiffies(delay));

}

static int cm3623_input_device_init(struct capella_cm3623_data *data, char *sensor_name)
{
    struct input_dev *dev;
    int err;

    printk(KERN_INFO "[CM3623] : %s\n", __func__);
    dev = input_allocate_device();
    if (!dev) {
        printk(KERN_INFO "[CM3623] : input_allocate_device fail\n");
        return -ENOMEM;
    }

    dev->name = sensor_name;
    dev->id.bustype = BUS_I2C;
    dev->dev.parent = &data->client->dev;

    //TODO Maybe need to setup diff param by sensor name
    //set_bit(EV_ABS, dev->evbit); or dev->evbit[0] = BIT_MASK(EV_ABS);
    if (!strcmp(sensor_name, "light")) {
        input_set_abs_params(dev, ABS_X , 0, 65536 , 0, 0);
    }
    else if (!strcmp(sensor_name, "proximity")) {
        input_set_abs_params(dev, ABS_X , 0, 65536 , 0, 0);
    }
    input_set_capability(dev, EV_ABS, ABS_X);
    input_set_drvdata(dev, data);


    err = input_register_device(dev);
    if (err < 0) {
        input_free_device(dev);
        printk(KERN_INFO "[cm3623] : Init input device fail\n");
        return err;
    }

    if (!strcmp(sensor_name, "light"))   data->als_input_dev = dev;
    else if (!strcmp(sensor_name, "proximity")) data->ps_input_dev = dev;
    D("[CM3623] : %s [%s] successful\n", __func__, sensor_name);

    return 0;

}

static int cm3623_init_device(struct i2c_client *client)
{
#if CM3623_USE_INTERRUPT
    //TODO to check interrupt pin state
#endif
    // cm3623 mode setting
    char reg_buff[1] = {0};
    int i, rs = 0;

    printk(KERN_INFO "[CM3623] : %s\n", __func__);
    client->addr = CM3623_PS_I2C_ADDR;
    reg_buff[0] = CM3623_PROXIMITY_OFF;     // turn off proximity sensor
    rs = i2c_write(client, reg_buff, 1);
    if (rs != 0) {
        printk(KERN_INFO "[CM3623] : i2c_write reg 0xB0 fail\n");
        return -1;
    }

    rs = gpio_request(CM3623_INT, "cm3623_int");

    rs = gpio_direction_input(CM3623_INT);

    for (i=0; i<3; i++) {
        msleep(2);
        rs = gpio_get_value(CM3623_INT);
        if (rs) break;
    }
    if (!rs) return -1;

    client->addr =  CM3623_INITIAL_I2C_ADDR;
    reg_buff[0] = CM3623_INIT_VALUE;
    rs = i2c_write(client, reg_buff, 1);
    if (rs != 0) {
        printk(KERN_INFO "[CM3623] : i2c_write reg 0x22 fail\n");
        return -1;
    }

    // init value to 0x20 ALS
    client->addr = CM3623_ALS_I2C_ADDR;
    reg_buff[0] = CM3623_LIGHT_OFF;   // turn off ambient light sensor
    rs = i2c_write(client, reg_buff, 1);
    if (rs != 0) {
        printk(KERN_INFO "[CM3623] : i2c_write reg 0x20 fail\n");
        return -1;
    }

    // init value to 0xB2 PS THD
    client->addr = CM3623_I2C_THD_ADDR;
    reg_buff[0] = 0x3;                  // 0x3 is min ; threshold if ps value >= THD , LIGHT_OUT_INT pin from high to low.
    rs = i2c_write(client, reg_buff, 1);
    if (rs != 0) {
        printk(KERN_INFO "[CM3623] : i2c_write reg 0xB2 fail\n");
        return -1;
    }

    // init value to 0xB0 PS
    client->addr = CM3623_PS_I2C_ADDR;
    reg_buff[0] = CM3623_PROXIMITY_OFF;     // turn off proximity sensor
    rs = i2c_write(client, reg_buff, 1);
    if (rs != 0) {
        printk(KERN_INFO "[CM3623] : i2c_write reg 0xB0 fail\n");
        return -1;
    }
    D("[CM3623] : Finish %s\n", __func__);
    D("-------------------------------\n");
    return 0;
}

// For PS attribute definition and call_back function
// [START]
//TODO For PS attribute call_back function
static int cm3623_ps_get_enable(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);

    D("[CM3623] : %s\n", __func__);
    return atomic_read(&cm3623->ps_enable);
}

static void cm3623_ps_set_enable(struct device *dev, int enable)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);
    unsigned int delay;
    char reg_buff[1] = {0};
    int rs = 0;

    D("[CM3623] : %s\n", __func__);
    delay = atomic_read(&cm3623->ps_delay);

    mutex_lock(&cm3623->enable_mutex);
    if (enable == 1) {                              //enable if state will be changed
        if (!atomic_cmpxchg(&cm3623->ps_enable, 0, 1)) {
            cm3623_proximity_status = true;

            cm3623->client->addr = CM3623_PS_I2C_ADDR;
            reg_buff[0] = CM3623_PROXIMITY_ON;      // turn on proximity sensor
            rs = i2c_write(cm3623->client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : i2c_write reg 0xB0 fail\n");
            }
            else {
                atomic_set(&cm3623->ps_enable, enable);
                schedule_delayed_work(&cm3623->ps_work, msecs_to_jiffies(delay));
                D("[CM3623] : ps_set_enable = 1 => [SUCCEED]\n");
            }
        }
    }
    else if (enable == 0) {                         //disable if state will be changed
        if (atomic_cmpxchg(&cm3623->ps_enable, 1, 0)) {
            cm3623_proximity_status = false;
            cancel_delayed_work_sync(&cm3623->ps_work);
            atomic_set(&cm3623->ps_enable, enable);

            cm3623->client->addr = CM3623_PS_I2C_ADDR;
            reg_buff[0] = CM3623_PROXIMITY_OFF;     // turn off proximity sensor
            rs = i2c_write(cm3623->client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : i2c_write reg 0xB0 fail\n");
            }
            else {
                D("[CM3623] : ps_set_enable = 0 => [SUCCEED]\n");
            }
        }
    }
    mutex_unlock(&cm3623->enable_mutex);
}

static int cm3623_ps_get_delay(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);
    
    D("[CM3623] : %s\n", __func__);
    return atomic_read(&cm3623->ps_delay);
}

static void cm3623_ps_set_delay(struct device *dev, int delay)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client); 
    
    D("[CM3623] : %s\n", __func__);
    atomic_set(&cm3623->ps_delay, delay);
}


/*
 * sysfs device attributes
 */
static ssize_t cm3623_ps_enable_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    D("[CM3623] : %s\n", __func__);
    return sprintf(buf, "%d\n", cm3623_ps_get_enable(dev));
}

static ssize_t cm3623_ps_enable_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    unsigned long enable;
    
    D("[CM3623] : %s\n", __func__);
    enable  = simple_strtoul(buf, NULL, 10);
    if ((enable == 0) || (enable == 1)) {
        cm3623_ps_set_enable(dev, enable);
    }

    return count;
}

static ssize_t cm3623_ps_delay_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", cm3623_ps_get_delay(dev));
}

static ssize_t cm3623_ps_delay_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    unsigned long delay;
    
    D("[CM3623] : %s\n", __func__);
    delay  = simple_strtoul(buf, NULL, 10);
    cm3623_ps_set_delay(dev, delay);

    return count;
}

static ssize_t cm3623_ps_data_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct capella_cm3623_data *cm3623 = input_get_drvdata(input);
    int ps_data=0;

    mutex_lock(&cm3623->data_mutex);
    ps_data = cm3623->ps_last;
    mutex_unlock(&cm3623->data_mutex);
    return sprintf(buf, "%d\n", ps_data);       
}

//TODO For PS attribute definition
/* ATTR DEFINE
static DEVICE_ATTR(delay, S_IRUGO|S_IWUSR|S_IWGRP, cm3623_ps_delay_show, cm3623_ps_delay_store);
static DEVICE_ATTR(enable, S_IRUGO|S_IWUSR|S_IWGRP, cm3623_ps_enable_show, cm3623_ps_enable_store);
static DEVICE_ATTR(wake, S_IWUSR|S_IWGRP, NULL, sensor_wake_store);
static DEVICE_ATTR(data, S_IRUGO, cm3623_ps_data_show, NULL);
static DEVICE_ATTR(status, S_IRUGO, sensor_status_show, NULL);
## DEVICE_ATTR(name, mode, show, store)
*/
static struct device_attribute dev_attr_delay_ps = {
    .attr = {
        .name = "delay",
        .mode = S_IWUSR | S_IRUGO,
    },
    .show = cm3623_ps_delay_show,
    .store = cm3623_ps_delay_store,
};

static struct device_attribute dev_attr_enable_ps = {
    .attr = {
        .name = "enable",
        .mode = S_IRUGO|S_IWUSR|S_IWGRP,
    },
    .show =cm3623_ps_enable_show,
    .store = cm3623_ps_enable_store,
};

static struct device_attribute dev_attr_data_ps = {
    .attr = {
        .name = "data",
        .mode = S_IRUGO,
    },
    .show =cm3623_ps_data_show,
};

static struct attribute *cm3623_ps_attributes[] = {
    &dev_attr_delay_ps.attr,
    &dev_attr_enable_ps.attr,
//  &dev_attr_wake.attr,
    &dev_attr_data_ps.attr,
//  &dev_attr_status.attr,
    NULL
};

static struct attribute_group cm3623_ps_attribute_group = {
    .attrs = cm3623_ps_attributes
};
// [END]

// For ALS attribute definition and call_back function
// [START]
//TODO For ALS attribute call_back function
static int cm3623_als_get_enable(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);
    
    D("[CM3623] : %s\n", __func__);
    return atomic_read(&cm3623->als_enable);
}

static void cm3623_als_set_enable(struct device *dev, int enable)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);
    unsigned int delay;
    char reg_buff[1] = {0};
    int rs = 0;

    D("[CM3623] : %s\n", __func__);
    delay = atomic_read(&cm3623->als_delay);

    mutex_lock(&cm3623->enable_mutex);
    //enable if state will be changed
    if (enable == 1) {
        if (!atomic_cmpxchg(&cm3623->als_enable, 0, 1)) {
            cm3623_light_status = true;
            cm3623->client->addr = CM3623_ALS_I2C_ADDR;
            reg_buff[0] = CM3623_LIGHT_ON;   // turn on ambient light sensor
            rs = i2c_write(cm3623->client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : Turn on light sensor setting => [FAIL]\n");
            }
            else {
                atomic_set(&cm3623->als_enable, enable);
                schedule_delayed_work(&cm3623->als_work, msecs_to_jiffies(delay));
                D("[CM3623] : als_set_enable = 1 => [SUCCEED]\n");
            }
        }
    }
    //disable if state will be changed
    else if (enable == 0) {
        if (atomic_cmpxchg(&cm3623->als_enable, 1, 0)) {
            cm3623_light_status = false;
            cancel_delayed_work_sync(&cm3623->als_work);
            atomic_set(&cm3623->als_enable, enable);

            cm3623->client->addr = CM3623_ALS_I2C_ADDR;
            reg_buff[0] = CM3623_LIGHT_OFF;   // turn off ambient light sensor
            rs = i2c_write(cm3623->client, reg_buff, 1);
            if (rs != 0) {
                printk(KERN_INFO "[CM3623] : Turn off light sensor setting => [FAIL]\n");
            }
            else {
                D("[CM3623] : als_set_enable = 0 => [SUCCEED]\n");
            }
        }
    }
    mutex_unlock(&cm3623->enable_mutex);
}

static int cm3623_als_get_delay(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);

    return atomic_read(&cm3623->als_delay);
}

static void cm3623_als_set_delay(struct device *dev, int delay)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client); 
    
    atomic_set(&cm3623->als_delay, delay);
}



/*
 * sysfs device attributes
 */

static ssize_t cm3623_als_enable_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    D("[CM3623] : %s\n", __func__);
    return sprintf(buf, "%d\n", cm3623_als_get_enable(dev));
}

static ssize_t cm3623_als_enable_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    unsigned long enable;

    D("[CM3623] : %s\n", __func__);
    enable = simple_strtoul(buf, NULL, 10);
    if ((enable == 0) || (enable == 1)) {
        cm3623_als_set_enable(dev, enable);
    }

    return count;
}

static ssize_t cm3623_als_delay_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    D("[CM3623] : %s\n", __func__);
    return sprintf(buf, "%d\n", cm3623_als_get_delay(dev));
}

static ssize_t cm3623_als_delay_store(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
    unsigned long delay;

    D("[CM3623] : %s\n", __func__);
    // long simple_strtoul(const char *cp, char **endp, unsigned int base
    // Transfer string to long int by base
    delay  = simple_strtoul(buf, NULL, 10);     
    cm3623_als_set_delay(dev, delay);

    return count;
}

static ssize_t cm3623_als_data_show(struct device *dev, struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct capella_cm3623_data *cm3623 = input_get_drvdata(input);
    int als_data=0;
    
    D("[CM3623] : %s\n", __func__);
    mutex_lock(&cm3623->data_mutex);
    als_data =cm3623->als_last;
    mutex_unlock(&cm3623->data_mutex);
    return sprintf(buf, "%d\n", als_data);       
}

// For ALS attribute definition
static DEVICE_ATTR(delay, S_IRUGO|S_IWUSR|S_IWGRP, cm3623_als_delay_show, cm3623_als_delay_store);
static DEVICE_ATTR(enable, S_IRUGO|S_IWUSR|S_IWGRP, cm3623_als_enable_show, cm3623_als_enable_store);
//static DEVICE_ATTR(wake, S_IWUSR|S_IWGRP, NULL, sensor_wake_store);
static DEVICE_ATTR(data, S_IRUGO, cm3623_als_data_show, NULL);
//static DEVICE_ATTR(status, S_IRUGO, sensor_status_show, NULL);

static struct attribute *cm3623_als_attributes[] = {
    &dev_attr_delay.attr,
    &dev_attr_enable.attr,
//  &dev_attr_wake.attr,
    &dev_attr_data.attr,
//  &dev_attr_status.attr,
    NULL
};

static struct attribute_group cm3623_als_attribute_group = {
    .attrs = cm3623_als_attributes
};
// [END]


//TODO
static int cm3623_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    int res = 0;

    printk(KERN_INFO "[CM3623] : %s\n", __func__);
    
    /* [START] allocate struct capella_cm3623_data *this ; and initial device as i2c client */
    this_data = kzalloc(sizeof(struct capella_cm3623_data), GFP_KERNEL);
    if (!this_data) {
        res = -ENOMEM;
        goto err_0;
    }
    // init als mutex
    mutex_init(&this_data->enable_mutex);
    mutex_init(&this_data->data_mutex);
    
    /* check i2c functionality is workable */
    /* Suppose to be I2C_FUNC_I2C */
    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
        printk(KERN_INFO "[cm3623] : i2c_check_functionality[I2C_FUNC_I2C] error!\n");
        res = -ENOTSUPP;
        goto err_1;
    }

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_SMBUS_BYTE)) {
        printk(KERN_INFO "[cm3623] : i2c_check_functionality[I2C_FUNC_SMBUS_BYTE] error!\n");
        res = -ENOTSUPP;
        goto err_1;
    }
    
    //strlcpy(client->name, ISL29018_DRIVER_NAME, I2C_NAME_SIZE);
    i2c_set_clientdata(client, this_data);
    this_data->client = client;
    /* [END] */

    /* TODO [START] : Initialize cm3623 chip */
    res = cm3623_init_device(client);
    if (res < 0) {
        goto err_2;
    }
    /* [END] */

    /* [START] Init input device for ALS and PS */
    // For ALS
    res = cm3623_input_device_init(this_data, "light");
    if (res < 0) {
        goto err_4;
    }
    atomic_set(&this_data->als_delay, CM3623_DEFAULT_DELAY); //TODO: depends on cm3623 datasheet

    /* register misc device */
    //TODO: nessesary or not
    /*
    res = misc_register(&cm3623_als_dev);
    if (res < 0) {
        pr_err("[CM3623] : cm3623_dev register failed! error code:[%d]\n", res);
        goto out7;
    }
    */

    res = sysfs_create_group(&this_data->als_input_dev->dev.kobj, &cm3623_als_attribute_group);
    if (res < 0) {
        input_unregister_device(this_data->ps_input_dev);
        input_free_device(this_data->ps_input_dev);
        return res;
    }    

    INIT_DELAYED_WORK(&this_data->als_work, als_work_func);

    // For PS
    res = cm3623_input_device_init(this_data, "proximity");
    if (res < 0) {
        goto err_5;
    }
    atomic_set(&this_data->ps_delay, CM3623_DEFAULT_DELAY); //TODO: depends on cm3623 datasheet
    
    /* register misc device */
    //TODO: necessary or not
    /*
    res = misc_register(&cm3623_ps_dev);
    if (res < 0) {
        pr_err("[CM3623] : cm3623_dev register failed! error code:[%d]\n", res);
        goto out4;
    }
    */
    
    res = sysfs_create_group(&this_data->ps_input_dev->dev.kobj, &cm3623_ps_attribute_group);
    if (res < 0) {
        input_unregister_device(this_data->ps_input_dev);
        input_free_device(this_data->ps_input_dev);
        return res;
    }

    INIT_DELAYED_WORK(&this_data->ps_work, ps_work_func);

#ifdef CONFIG_HAS_EARLYSUSPEND
    early_suspend.level = EARLY_SUSPEND_LEVEL_STOP_DRAWING + 1;
    early_suspend.suspend = cm3623_light_early_suspend;
    early_suspend.resume = cm3623_light_late_resume;
    register_early_suspend(&early_suspend);
#endif

    return 0;

err_0:
    printk(KERN_INFO "[cm3623] : Setup private data fail\n");
    return res;

err_1:
    printk(KERN_INFO "[cm3623] : i2c check fail\n");
    kfree(this_data);
    return res;

err_2:
    printk(KERN_INFO "[cm3623] : Initialize the cm3623 chip fail\n");
    kfree(this_data);
    return res;

err_4:
    printk(KERN_INFO "[cm3623] : Set cm3623 ALS input device fail\n");
    kfree(this_data);
    return res;
   
err_5:
    printk(KERN_INFO "[cm3623] : Set cm3623 PS input device fail\n");
    kfree(this_data);
    return res;
}

static int cm3623_remove(struct i2c_client *client)
{
    struct capella_cm3623_data *cm3623 = i2c_get_clientdata(client);

    input_unregister_device(cm3623->als_input_dev);
    input_unregister_device(cm3623->ps_input_dev);
    kfree(cm3623);

    return 0;
}

#define CAPELLA_CM3623 "capella_cm3623"
static const struct i2c_device_id cm3623_id[] = {
    {CAPELLA_CM3623, 0},
    {},
};

MODULE_DEVICE_TABLE(i2c, cm3623_id);

static struct i2c_driver cm3623_driver = {
    .driver = {
        .name = CAPELLA_CM3623,
        .owner = THIS_MODULE,
    },
    .probe = cm3623_probe,
    .remove = cm3623_remove,
    .suspend = cm3623_suspend,
    .resume = cm3623_resume,
    .id_table = cm3623_id,
};

static int __init capella_cm3623_init(void)
{
    int ret, C6_hw_id;
    struct vreg *vreg_rftx = NULL;
    cci_smem_value_t *smem_cci_smem_value = 0;
    unsigned char LS_PS_NV_Value[4];
    uint16_t light_temp, proximity_temp;

    printk(KERN_INFO "[CM3623] : %s\n", __func__);
    smem_cci_smem_value = smem_alloc( SMEM_CCI_SMEM_VALUE, sizeof( cci_smem_value_t ));
    C6_hw_id = smem_cci_smem_value->cci_hw_id;

    switch (C6_hw_id)
    {
        case HW_ID_DVT1:
            break;
        case HW_ID_DVT1_2:
        case HW_ID_DVT2:
        case HW_ID_DVT2_2:
        default:
            printk(KERN_INFO "[CM3623] : HW_ID is HW_ID_DVT1_2\n");
            vreg_rftx = vreg_get(0, "rftx");
            ret = vreg_set_level(vreg_rftx, 3000);
            if (ret) {
                printk(KERN_INFO "[CM3623] : vreg set rftx level failed\n");
            }
            mdelay(2);
            ret = vreg_enable(vreg_rftx);
            if (ret) {
                printk(KERN_INFO "[CM3623] : vreg enable rftx failed\n");
            }
            mdelay(10);
            printk(KERN_INFO "[CM3623] : vreg enable rftx => [succeed]\n");
            break;
    }

    memcpy(LS_PS_NV_Value, &(smem_cci_smem_value ->cci_smem_define_item[CCI_SMEM_ITEM_0]), CCI_SMEM_ITEM_0_SIZE);
    light_temp = (LS_PS_NV_Value[2]<<8) | LS_PS_NV_Value[3];
    proximity_temp = (LS_PS_NV_Value[0]<<8) | LS_PS_NV_Value[1];

    if ((light_temp < LIGHT_NV_ITEM_VALUE_HIGH) && (light_temp > LIGHT_NV_ITEM_VALUE_LOW)) LIGHT_NV_ITEM_VALUE = light_temp;
    if (proximity_temp && (proximity_temp < PROXIMITY_NV_ITEM_VALUE_HIGH)) PROXIMITY_NV_ITEM_VALUE = proximity_temp;
    printk("Light Sensor Bench Level = %d , Proximity Sensor Bench Level = %d\n", LIGHT_NV_ITEM_VALUE, PROXIMITY_NV_ITEM_VALUE);

    ret = i2c_add_driver(&cm3623_driver);

    if (ret != 0) {
        printk(KERN_ERR "[%s] : can't add i2c driver\n", __func__);
        return ret;
    }
    return ret;
}
module_init(capella_cm3623_init);

static void __exit capella_cm3623_exit(void)
{
    i2c_del_driver(&cm3623_driver);
}
module_exit(capella_cm3623_exit);

MODULE_DESCRIPTION("CAPELLA CM3623 light and proximity driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);
