/*
 * Copyright (C) 2010 Trusted Logic S.A.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/i2c.h>
#include <linux/irq.h>
#include <linux/jiffies.h>
#include <linux/uaccess.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/platform_device.h>
#include <linux/gpio.h>
#include <linux/miscdevice.h>
#include <linux/spinlock.h>
#include <linux/pn544.h>

#define MAX_BUFFER_SIZE	512
#define TAG "[NFC]"
#define LOG(s, args...) printk("[NFC] " s "\r\n", ## args)

struct pn544_dev	{
	wait_queue_head_t	read_wq;
	struct mutex		read_mutex;
	struct i2c_client	*client;
	struct miscdevice	pn544_device;
	unsigned int 		ven_gpio;
	unsigned int 		firm_gpio;
	unsigned int		irq_gpio;
	unsigned int		ldo_gpio; //used from DVT1_2
	bool			irq_enabled;
	spinlock_t		irq_enabled_lock;
	int irq_num;
};

/* Add NFC power control by Arlin (ref DA80)*/
#ifdef CONFIG_CCI_HWID
#include <mach/../../smd_private.h>
#include "mach/../../cci_smem.h"
cci_smem_value_t *smem_cci_smem_value;
int hw_id;
#endif

#include <mach/vreg.h>

int pwr_state = 0;

static bool pn544_enable_irq(struct pn544_dev *pn544_dev, bool enable);

static void pn544_ven(unsigned int ven, int is_on)
{
	#ifdef CONFIG_CCI_HWID
		gpio_direction_output(ven, is_on);

		if((hw_id == HW_ID_EVT0) || (hw_id == HW_ID_DVT1) || (hw_id == HW_ID_DVT1_2)){
			msleep(10);
		}else if(hw_id == HW_ID_DVT2){
			if(is_on)
				msleep(30);
			else
				msleep(55);
		}else{ //DVT2_2
			if(is_on)
				msleep(17);
			else
				msleep(30);
		}
	#else
		gpio_set_value(ven, is_on);
		msleep(10);
	#endif
}

static int pn544_power(struct pn544_dev *pn544_dev, int on)
{
    static struct vreg *vreg_nfc;
    int ret = 0;

	#ifdef CONFIG_CCI_HWID
	if((hw_id == HW_ID_EVT0) ||(hw_id == HW_ID_DVT1)){ //before DVT1_2
		vreg_nfc = vreg_get(0, "rftx");
		if (IS_ERR(vreg_nfc)) {
			LOG("vreg_get(rftx) vreg_nfc fail.");
			goto error;
		}
		LOG("vreg_get(rftx) vreg_nfc ok");
	}
	#endif
 
    if(on > 0)
    {
	#ifdef CONFIG_CCI_HWID
	if((hw_id == HW_ID_EVT0) ||(hw_id == HW_ID_DVT1)){//before DVT1_2
		ret = vreg_set_level(vreg_nfc, 1800);
		if (ret) {
			LOG("vreg_set_level(rftx) vreg_nfc fail. ret = %d", ret);
			vreg_put(vreg_nfc);
			goto error;
		}

	        ret = vreg_enable(vreg_nfc);
	        if (ret) {
			LOG("vreg_enable(rftx) vreg_nfc fail");
			vreg_put(vreg_nfc);
			goto error;
	        }
		LOG("vreg_enable(rftx) vreg_nfc 1.8V ok");
	}else if(hw_id == HW_ID_DVT2) {
		if(gpio_is_valid(pn544_dev->ldo_gpio)){
			gpio_direction_output(pn544_dev->ldo_gpio, 1);
			LOG("ldo_gpio turn on ok");
			msleep(30);
		}
	}
	#endif

	if(on == 2){
		LOG("%s: power on with firmware download (requires hw reset)", __func__);

		pn544_ven(pn544_dev->ven_gpio, 1);
		gpio_set_value(pn544_dev->firm_gpio, 1);
		msleep(10);
		pn544_ven(pn544_dev->ven_gpio, 0);
		pn544_ven(pn544_dev->ven_gpio, 1);
	}else{ /*just power on*/
		gpio_set_value(pn544_dev->firm_gpio, 0);
		pn544_ven(pn544_dev->ven_gpio, 1);
		
		/*reset twice*/
		pn544_ven(pn544_dev->ven_gpio, 0);
		pn544_ven(pn544_dev->ven_gpio, 1);
		pn544_ven(pn544_dev->ven_gpio, 0);
		pn544_ven(pn544_dev->ven_gpio, 1);
		LOG("%s: ON", __func__);
	}
    }
    else /*power off */
    {
    	gpio_set_value(pn544_dev->firm_gpio, 0);
	pn544_ven(pn544_dev->ven_gpio, 0);

	#ifdef CONFIG_CCI_HWID
	if((hw_id == HW_ID_EVT0) ||(hw_id == HW_ID_DVT1)){	//before DVT1_2
		ret = vreg_disable(vreg_nfc);
		if (ret) {
			LOG("vreg_disable(rftx) vreg_nfc fail");
			vreg_put(vreg_nfc);
			goto error;
	        }
	        LOG("vreg_disable(rftx) vreg_nfc ok");
	}else{
		if(gpio_is_valid(pn544_dev->ldo_gpio)){
			gpio_direction_output(pn544_dev->ldo_gpio, 0);
			LOG("ldo_gpio turn off ok");
		}
	}
	#endif
	LOG("%s: OFF", __func__);
    }

    return 0;

error:
    return -1;
}

static bool pn544_enable_irq(struct pn544_dev *pn544_dev, bool enable)
{
	unsigned long flags;
	bool ret;

	spin_lock_irqsave(&pn544_dev->irq_enabled_lock, flags);

	ret = pn544_dev->irq_enabled;
	if (enable != pn544_dev->irq_enabled)
	{
		if (enable)
			enable_irq(pn544_dev->irq_num);
		else
			disable_irq_nosync(pn544_dev->irq_num);

		pn544_dev->irq_enabled = enable;
	}

	spin_unlock_irqrestore(&pn544_dev->irq_enabled_lock, flags);
	return ret;
}

static irqreturn_t pn544_dev_irq_handler(int irq, void *dev_id)
{
	struct pn544_dev *pn544_dev = dev_id;

	pn544_enable_irq(pn544_dev, false);

	/* Wake up waiting readers */
	if (pwr_state > 0)
		wake_up(&pn544_dev->read_wq);

	return IRQ_HANDLED;
}

static ssize_t pn544_dev_read(struct file *filp, char __user *buf,
		size_t count, loff_t *offset)
{
	struct pn544_dev *pn544_dev = filp->private_data;
	char tmp[MAX_BUFFER_SIZE];
	int ret;

	if (count > MAX_BUFFER_SIZE)
		count = MAX_BUFFER_SIZE;

	pr_debug("%s %s: reading %zu bytes.\n", TAG,  __func__, count);

	mutex_lock(&pn544_dev->read_mutex);

	if (!gpio_get_value(pn544_dev->irq_gpio) || (pwr_state == 0)) {
		if (filp->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			LOG("%s: get irq value fail, returned %d", __func__, ret);
			goto fail;
		}

		pn544_enable_irq(pn544_dev, true);
		ret = wait_event_interruptible(pn544_dev->read_wq,
				gpio_get_value(pn544_dev->irq_gpio) && (pwr_state != 0));

		pn544_enable_irq(pn544_dev, false);

		if (ret){
			LOG("%s: wait_event_interrupt returned %d", __func__, ret);
			goto fail;
		}
	}

	/* Read data */
	ret = i2c_master_recv(pn544_dev->client, tmp, count);
	mutex_unlock(&pn544_dev->read_mutex);

	if (ret < 0) {
		pr_err("%s %s: i2c_master_recv returned %d\n", TAG, __func__, ret);
		return ret;
	}
	if (ret > count) {
		pr_err("%s %s: received too many bytes from i2c (%d)\n",
			TAG, __func__, ret);
		return -EIO;
	}

	if (copy_to_user(buf, tmp, ret)) {
		LOG("%s: failed to copy to user space", __func__);
		return -EFAULT;
	}
	return ret;

fail:
	mutex_unlock(&pn544_dev->read_mutex);
	return ret;
}

static ssize_t pn544_dev_write(struct file *filp, const char __user *buf,
		size_t count, loff_t *offset)
{
	struct pn544_dev  *pn544_dev;
	char tmp[MAX_BUFFER_SIZE];
	int ret;
	int i = 0;
	 u8 *data;
	 int retry = 0;

	pn544_dev = filp->private_data;

	if (count > MAX_BUFFER_SIZE)
		count = MAX_BUFFER_SIZE;

	if (copy_from_user(tmp, buf, count)) {
		pr_err("%s %s: failed to copy from user space\n", TAG, __func__);
		return -EFAULT;
	}

	/* Write data */
	do{
		ret = i2c_master_send(pn544_dev->client, tmp, count);
		if (ret != count) {
			pr_err("%s %s: (%d) i2c_master_send returned %d\n", TAG, __func__, retry+1, ret);
			ret = -EIO;

			/*i2c debug log*/
			data = tmp;
			printk("%s %s: i2c error data ", TAG, __func__);
			for ( i=0 ; i < count ;i++) {
				printk("[%d]:0x%x, ", i, *(data+i));
			}
			printk(".\n");
			msleep(10);
		}else
			break;
	}while(retry++ < 2);
	return ret;
}

static int pn544_dev_open(struct inode *inode, struct file *filp)
{
	struct pn544_dev *pn544_dev = container_of(filp->private_data,
						struct pn544_dev,
						pn544_device);

	filp->private_data = pn544_dev;
	
	LOG("%s: pn544_dev[%p]", __func__, pn544_dev);
	pr_debug("%s %s: %d,%d\n", TAG, __func__, imajor(inode), iminor(inode));

	return 0;
}

static int pn544_dev_ioctl(struct inode *inode, struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct pn544_dev *pn544_dev = filp->private_data;
	int ret;

	LOG("%s: cmd = %d, arg = %ld", __func__, cmd, arg);
	if (filp->f_flags & O_NONBLOCK) {
		ret = -EAGAIN;
		LOG("%s: handle with nonblocking IO, returned %d", __func__, ret);
		return ret;
	}

	switch (cmd) {
	case PN544_SET_PWR:
		if (arg >= 0) {
			pn544_enable_irq(pn544_dev, false);
			if (arg != pwr_state){
				pwr_state = 0;
				pn544_power(pn544_dev, arg);
			}
			msleep(10);

			if (arg > 0)
				pn544_enable_irq(pn544_dev, true);
			pwr_state = arg;
		} else {
			pr_err("%s %s: bad arg %u\n", TAG, __func__, (unsigned int)arg);
			return -EINVAL;
		}
		break;
	default:
		pr_err("%s %s: bad ioctl %u\n", TAG, __func__, cmd);
		return -EINVAL;
	}

	return 0;
}

static const struct file_operations pn544_dev_fops = {
	.owner	= THIS_MODULE,
	.llseek	= no_llseek,
	.read	= pn544_dev_read,
	.write	= pn544_dev_write,
	.open	= pn544_dev_open,
	.ioctl  = pn544_dev_ioctl,
};

static int pn544_probe(struct i2c_client *client,
		const struct i2c_device_id *id)
{
	int ret;
	struct pn544_i2c_platform_data *platform_data;
	struct pn544_dev *pn544_dev;

	platform_data = client->dev.platform_data;

	#ifdef CONFIG_CCI_HWID
	if((hw_id == HW_ID_EVT0) ||(hw_id == HW_ID_DVT1)) //before DVT1_2
		LOG("%s: irq_gpio[%d], ven_gpio[%d], firm_gpio[%d]", __func__,
			platform_data->irq_gpio, platform_data->ven_gpio, platform_data->firm_gpio);
	else
		LOG("%s: irq_gpio[%d], ven_gpio[%d], firm_gpio[%d], ldo_gpio[%d]", __func__,
			platform_data->irq_gpio, platform_data->ven_gpio, platform_data->firm_gpio, platform_data->ldo_gpio);
	#endif

	if (platform_data == NULL) {
		pr_err("%s %s: probe fail\n", TAG, __func__);
		return  -ENODEV;
	}

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		pr_err("%s %s : need I2C_FUNC_I2C\n", TAG, __func__);
		return  -ENODEV;
	}

	ret = gpio_request(platform_data->irq_gpio, "nfc_int");
	if (ret)
		return  -ENODEV;
	ret = gpio_request(platform_data->ven_gpio, "nfc_ven");
	if (ret)
		goto err_ven;
	ret = gpio_request(platform_data->firm_gpio, "nfc_firm");
	if (ret)
		goto err_firm;
	#ifdef CONFIG_CCI_HWID
	if((hw_id != HW_ID_EVT0) && (hw_id != HW_ID_DVT1)){ //from DVT1_2
		if (gpio_request(platform_data->ldo_gpio, "nfc_ldo")){
			LOG("%s: ldo_gpio request error", __func__);
			goto err_ldo;
		}
	}
	#endif

	pn544_dev = kzalloc(sizeof(*pn544_dev), GFP_KERNEL);
	if (pn544_dev == NULL) {
		dev_err(&client->dev,
				"%s failed to allocate memory for module data\n", TAG);
		ret = -ENOMEM;
		goto err_exit;
	}

	pn544_dev->irq_gpio = platform_data->irq_gpio;
	pn544_dev->ven_gpio  = platform_data->ven_gpio;
	pn544_dev->firm_gpio  = platform_data->firm_gpio;
	#ifdef CONFIG_CCI_HWID
	if((hw_id != HW_ID_EVT0) && (hw_id != HW_ID_DVT1)){ //from DVT1_2
		pn544_dev->ldo_gpio  = platform_data->ldo_gpio;
		if(hw_id != HW_ID_DVT1_2){
			pn544_dev->ven_gpio  = platform_data->ldo_gpio; //127 wiil replace 115 in DVT2, DVT2_2
		}
	}
	#endif
	pn544_dev->client   = client;

	/* init mutex and queues */
	init_waitqueue_head(&pn544_dev->read_wq);
	mutex_init(&pn544_dev->read_mutex);
	spin_lock_init(&pn544_dev->irq_enabled_lock);

	pn544_dev->pn544_device.minor = MISC_DYNAMIC_MINOR;
	pn544_dev->pn544_device.name = "pn544";
	pn544_dev->pn544_device.fops = &pn544_dev_fops;
	pn544_dev->pn544_device.parent = &client->dev;

	ret = misc_register(&pn544_dev->pn544_device);
	if (ret) {
		pr_err("%s %s: misc_register failed\n", TAG, __FILE__);
		goto err_misc_register;
	}

	/* request irq.  the irq is set whenever the chip has data available
	 * for reading.  it is cleared when all data has been read.
	 */
	pr_info("%s %s: requesting IRQ %d\n", TAG, __func__, pn544_dev->irq_gpio);
	pn544_dev->irq_enabled = true;
	pn544_dev->irq_num = MSM_GPIO_TO_INT(pn544_dev->irq_gpio);
	gpio_direction_input(pn544_dev->irq_gpio);
	ret = request_irq(pn544_dev->irq_num, pn544_dev_irq_handler,
			  IRQF_TRIGGER_HIGH, client->name, pn544_dev);
	if (ret) {
		dev_err(&client->dev, "%s request_irq failed\n", TAG);
		goto err_request_irq_failed;
	}
	pn544_enable_irq(pn544_dev, false);
	i2c_set_clientdata(client, pn544_dev);
	
	LOG("%s: completed", __func__);

	return 0;

err_request_irq_failed:
	misc_deregister(&pn544_dev->pn544_device);
err_misc_register:
	mutex_destroy(&pn544_dev->read_mutex);
	kfree(pn544_dev);
err_exit:
	#ifdef CONFIG_CCI_HWID
	if((hw_id != HW_ID_EVT0) && (hw_id != HW_ID_DVT1)) //from DVT1_2
		gpio_free(platform_data->ldo_gpio);
	#else
	gpio_free(platform_data->firm_gpio);
	#endif
	pwr_state = 0;
err_ldo:
	gpio_free(platform_data->firm_gpio);
	pwr_state = 0;
err_firm:
	gpio_free(platform_data->ven_gpio);
	pwr_state = 0;
err_ven:
	gpio_free(platform_data->irq_gpio);
	pwr_state = 0;
	return ret;
}

static int pn544_remove(struct i2c_client *client)
{
	struct pn544_dev *pn544_dev;
	
	LOG("%s", __func__);

	pn544_dev = i2c_get_clientdata(client);
	free_irq(pn544_dev->irq_gpio, pn544_dev);
	misc_deregister(&pn544_dev->pn544_device);
	mutex_destroy(&pn544_dev->read_mutex);
	gpio_free(pn544_dev->irq_gpio);
	gpio_free(pn544_dev->ven_gpio);
	gpio_free(pn544_dev->firm_gpio);
	#ifdef CONFIG_CCI_HWID
	if((hw_id != HW_ID_EVT0) && (hw_id != HW_ID_DVT1)) //from DVT1_2
		gpio_free(pn544_dev->ldo_gpio);
	#endif
	kfree(pn544_dev);
	pwr_state = 0;

	return 0;
}

static const struct i2c_device_id pn544_id[] = {
	{ "pn544", 0 },
	{ }
};

static struct i2c_driver pn544_driver = {
	.id_table	= pn544_id,
	.probe		= pn544_probe,
	.remove		= pn544_remove,
	.driver		= {
		.owner	= THIS_MODULE,
		.name	= "pn544",
	},
};

/*
 * module load/unload record keeping
 */

static int __init pn544_dev_init(void)
{
	#ifdef CONFIG_CCI_HWID
	smem_cci_smem_value = smem_alloc( SMEM_CCI_SMEM_VALUE, sizeof( cci_smem_value_t ));

	if(smem_cci_smem_value->NCF_id == NFC_DISABLE){
		LOG("%s: cci_nfc_id: 0x%x", __func__, smem_cci_smem_value->NCF_id);

		/*set all pins to output low*/
		if (!gpio_request(114, "nfc_int"))
			gpio_direction_output(114, 0);

		if (!gpio_request(115, "nfc_ven"))
			gpio_direction_output(115, 0);

		if (!gpio_request(121, "nfc_firm"))
			gpio_direction_output(121, 0);

		if (!gpio_request(127, "nfc_ldo"))
			gpio_direction_output(127, 0);

		LOG("%s: Set all pins to output low", __func__);

		if(gpio_is_valid(114)){
			LOG("free gpio 114");
			gpio_free(114);
		}
		if(gpio_is_valid(115)){
			LOG("free gpio 115");
			gpio_free(115);
		}
		if(gpio_is_valid(121)){
			LOG("free gpio 121");
			gpio_free(121);
		}
		if(gpio_is_valid(127)){
			LOG("free gpio 127");
			gpio_free(127);
		}
		return -1;
	}
	hw_id = smem_cci_smem_value->cci_hw_id;
	#endif

	LOG("%s: Loading pn544 driver", __func__);
	return i2c_add_driver(&pn544_driver);
}
module_init(pn544_dev_init);

static void __exit pn544_dev_exit(void)
{
	LOG("%s: Unloading pn544 driver", __func__);
	i2c_del_driver(&pn544_driver);
}
module_exit(pn544_dev_exit);

MODULE_AUTHOR("Sylvain Fonteneau");
MODULE_DESCRIPTION("NFC PN544 driver");
MODULE_LICENSE("GPL");
